from __future__ import annotations
from typing import Tuple

import re


class Callsign:
    main: str
    ssid: int
    stroke: str

    def __init__(self, callsign: str):
        pattern = r"(?P<main>[\w*]+)(-(?P<ssid1>\d+))?(\/(?P<stroke1>\w+))?((\/(?P<stroke2>\w+))(-(?P<ssid2>\d+)))?$"
        callsign = callsign.upper().strip()
        match = re.match(pattern, callsign)
        if not match:
            raise ValueError(f"{callsign} is not a valid callsign")
        self.main = match.group("main")
        self.ssid = 0
        self.stroke = ""
        for group_name in match.groupdict():
            if group_name.startswith("stroke") and match.group(group_name):
                self.stroke = match.group(group_name)
            if group_name.startswith("ssid") and match.group(group_name):
                self.ssid = int(match.group(group_name))

    def __repr__(self) -> str:
        r = self.main
        if self.stroke:
            r += f"/{self.stroke}"
        if self.ssid > 0:
            r += f"-{self.ssid}"
        return r

    def _ax25_max_len_6_callsign(self) -> str:
        if self.stroke and len(self.main + self.stroke) + 1 <= 6:
            return "%-6s" % f"{self.main}/{self.stroke}"
        elif len(self.main) <= 6:
            return "%-6s" % self.main
        raise ValueError(f"'{self.main}' will not fit into AX25 callsign field")

    def to_ax25_bytes(self, is_last: bool = False, is_command: bool = False) -> bytes:
        fitted = bytes(map(lambda x: x << 1, bytes(self._ax25_max_len_6_callsign(), "ascii")))
        byte7 = 0x60
        if is_command:
            byte7 |= 0x80
        if is_last:
            byte7 |= 0x01
        byte7 |= (self.ssid & 0x0f) << 1
        return fitted + bytes([byte7])

    @staticmethod
    def from_ax25_bytes(ax25_bytes: bytes) -> Tuple[Callsign, bool]:
        if len(ax25_bytes) < 7:
            raise ValueError(f"AX25 callsign consists of 7 bytes, not {len(ax25_bytes)}")
        cs = str(bytes(map(lambda x: x >> 1, ax25_bytes[:6])), "ascii").strip()
        ssid = (ax25_bytes[6] & 0x1e) >> 1
        return Callsign(f"{cs}-{ssid}"), bool(ax25_bytes[6] & 1)


def callsign_of(obj) -> Callsign:
    if type(obj) is Callsign:
        return obj
    if type(obj) is str:
        return Callsign(obj)
    if type(obj) is bytes:
        return Callsign(str(obj, "ascii"))
    raise NotImplementedError(f"Constructing a callsign from '{type(obj)}' object")
