from __future__ import annotations
from typing import List, Optional
from bitarray import bitarray
from .bitutils import reverse_bits, do_bit_stuffing
from .callsign import Callsign, callsign_of
from .fcs import Checksum

import struct


FLAG_SYMBOL = bitarray("01111110")


class AX25Packet:
    destination: Callsign
    source: Callsign
    path: List[Callsign]
    control: int
    protocol: int
    info: str

    def __init__(self, source, destination, path=None, info: str = None, control: int = 0, protocol: int = 0):
        self.destination = callsign_of(destination)
        self.source = callsign_of(source)
        self.path = path_of(path)
        self.control = control
        self.protocol = protocol
        self.info = info

    def __repr__(self) -> str:
        r = f"{self.source}>{self.destination}"
        if self.path:
            r += ","
            r += ",".join(map(repr, self.path))
        r += ":"
        if self.info:
            r += self.info
        return r

    def to_bytes(self) -> bytes:
        b = b""
        b += self.destination.to_ax25_bytes(is_last=False)
        b += self.source.to_ax25_bytes(is_last=not bool(self.path))
        if self.path:
            last = len(self.path) - 1
            for idx, call in enumerate(self.path):
                b += call.to_ax25_bytes(is_last=(idx == last))
        b += struct.pack("BB", self.control, self.protocol)
        if self.info:
            b += bytes(self.info, "ascii")

        fcs = Checksum().feed_bytes(b).get()
        fcs_l = fcs & 0xff
        fcs_h = (fcs >> 8) & 0xff
        b += struct.pack("BB", fcs_l, fcs_h)
        b = bytes(map(reverse_bits, b))
        return b

    @staticmethod
    def from_bytes(bytes_: bytes) -> Optional[AX25Packet]:
        if len(bytes_) < 18:
            return None

        ax = AX25Packet("NONE", "NONE")

        bytes_, received_fcs = bytes_[:-2], bytes_[-2:]
        calculated_fcs = Checksum().feed_bytes(bytes_).get()
        c_fcs_l = calculated_fcs & 0xff
        c_fcs_h = (calculated_fcs >> 8) & 0xff
        calculated_fcs = struct.pack("BB", c_fcs_l, c_fcs_h)
        if calculated_fcs != received_fcs:
            return None

        ax.destination, _ = Callsign.from_ax25_bytes(bytes_[:7])
        ax.source, is_last = Callsign.from_ax25_bytes(bytes_[7:14])
        bytes_ = bytes_[14:]

        path_calls = []
        while not is_last:
            call, is_last = Callsign.from_ax25_bytes(bytes_[:7])
            path_calls.append(call)
            bytes_ = bytes_[7:]
        ax.path = path_calls

        ax.control, ax.protocol = bytes_[0], bytes_[1]
        ax.info = str(bytes_[2:], "ascii")
        return ax

    def to_packet_bits(self, start_flags: int = 24, end_flags: int = 24) -> bitarray:
        b = self.to_bytes()
        ba = bitarray()
        ba.frombytes(b)
        ba = do_bit_stuffing(ba, break_symbol=1, break_every=5)
        return FLAG_SYMBOL * start_flags + ba + FLAG_SYMBOL * end_flags


def path_of(obj) -> List[Callsign]:
    if not obj:
        return []
    if type(obj) is list:
        return list(map(callsign_of, obj))
    if type(obj) is str:
        return list(map(callsign_of, obj.split(",")))
    raise NotImplementedError(f"Constructing a callsign path from '{type(obj)}' object")
